/*	$Id: mesh.c 2007/11/23 19:06:04 mallet $ */

/*
 * Copyright (c) 1997 LAAS/CNRS                       --  Thu Aug  3 2006
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "libimages3d.h"


/**********************************************************************
 *
 * Generate triangulation and save in OFF file
 *
 **********************************************************************/

#define THRESHMODEL 1

#define mini(x,y) (x<y)?x:y;
#define maxi(x,y) (x>y)?x:y;

#define Z1R 1.0
#define Z1G 0.5
#define Z1B 0.5

#define Z2R 0.9
#define Z2G 1.0
#define Z2B 0.2

int
write_data_im3d_off (DATA_IM3D* im3d,
		     char*         nom_fichier,
		     double ATHRESH,
		     double RTHRESH,
		     unsigned char*  q_im_gauche)
{
  DATA_PT3D*    point;
  int           nblig;
  int           nbcol;
  int           i,j, NMAX;
  double        px,py,pz;
  int           c,count;
  char          nom_complet[256];
  FILE*         fichier;

  /* Line parameters */
  int           *PTRI;
  double        *PCOL, *PLIG, *PX, *PY, *PZ;
  char          *PVALID;
  int           NbTri;

  /* Filter parameters */
  int           pp1,pp2,pp3;
  double        l1,l2,l3,heron,area,lmax;
  int           tricount;
  int           i1,i2,i3,j1,j2,j3,color;
  float         fcol;

  /* ------------------------------------------------------ */

  /* On dcode les paramtres passs */
  nblig = im3d->header.nblig;
  nbcol = im3d->header.nbcol;

  /*----------------------------------------*/
  /* Compute the size of the working arrays */
  /*----------------------------------------*/
  count = 0;
  for (j = 0; j < nbcol; j++) {
    point = im3d->points_tab + j;
    for (i = 0; i < nblig; i++) {
      if (point->state == GOOD_PT3D) count++;
      point += nbcol;
    }
  }

  NMAX = count + 2;
  if ((PX = (double *) malloc (NMAX*sizeof(double))) == NULL)
    return ERR_IM3D_MALLOC;
  if ((PY = (double *) malloc (NMAX*sizeof(double))) == NULL)
    return ERR_IM3D_MALLOC;
  if ((PZ = (double *) malloc (NMAX*sizeof(double))) == NULL)
    return ERR_IM3D_MALLOC;
  if ((PCOL = (double *) malloc (NMAX*sizeof(double))) == NULL)
    return ERR_IM3D_MALLOC;
  if ((PLIG = (double *) malloc (NMAX*sizeof(double))) == NULL)
    return ERR_IM3D_MALLOC;

  if ((PVALID = malloc (2*NMAX)) == NULL)
    return ERR_IM3D_MALLOC;

  /*-----------------------*/
  /*  Fill the arrays      */
  /*-----------------------*/

  count = 0;
  for (j = 0; j < nbcol; j++) {
    point = im3d->points_tab + j;
    for (i = 0; i < nblig; i++) {
      if (point->state == GOOD_PT3D) {
	px = point->coord_1 ;
	py = point->coord_2 ;
	pz = point->coord_3 ;
	PX[count] = px;
	PY[count] = py;
	PZ[count] = pz;
	PCOL[count] = (double)j;
	PLIG[count] = (double)i;
	count++;
	if (count>NMAX) {
	  printf(">>> Node count exceeded\n");
	  break;
	}
      }
      point += nbcol;
    }
  }
  PTRI = Delaunay_2d(&NbTri, PCOL, PLIG, count);
  printf(">>> Found %d triangles\n",NbTri);

  /*****************************/
  /* Filtering large triangles */
  /*****************************/
  tricount = 0;
  for (c=0;c<NbTri;c++) {
    pp1 = PTRI[c];
    pp2 = PTRI[NbTri+c];
    pp3 = PTRI[2*NbTri+c];

    l1 = sqrt((PX[pp1]-PX[pp2])*(PX[pp1]-PX[pp2]) +
	      (PY[pp1]-PY[pp2])*(PY[pp1]-PY[pp2])
	      + (PZ[pp1]-PZ[pp2])*(PZ[pp1]-PZ[pp2]));
    l2 = sqrt((PX[pp1]-PX[pp3])*(PX[pp1]-PX[pp3]) +
	      (PY[pp1]-PY[pp3])*(PY[pp1]-PY[pp3])
	      + (PZ[pp1]-PZ[pp3])*(PZ[pp1]-PZ[pp3]));
    l3 = sqrt((PX[pp3]-PX[pp2])*(PX[pp3]-PX[pp2]) +
	      (PY[pp3]-PY[pp2])*(PY[pp3]-PY[pp2])
	      + (PZ[pp3]-PZ[pp2])*(PZ[pp3]-PZ[pp2]));

    heron = (l1+l2+l3)/2.0;

    area  = sqrt(heron*(heron-l1)*(heron-l2)*(heron-l3));

    lmax = maxi(l1,l2);
    lmax = maxi(lmax,l3);
    if ((area>ATHRESH)||(lmax>RTHRESH))
      PVALID[c] = 0;
    else {
      PVALID[c] = 1;
      tricount++;
    }
  }
  printf(">>> Number of triangles after area filtering = %d\n",tricount);

  /* ---------------------- */
  /* Export OFF file  */
  /* ---------------------- */
  sprintf (nom_complet, "%s.off",nom_fichier);
  fichier = fopen (nom_complet, "w");
  if (fichier == NULL)
    return (ERR_IM3D_OPEN_ASCII_FILE);

  /* le mot clef */
  fprintf (fichier, "OFF\n");

  fprintf(fichier,"%d %d 0\n",count,tricount);
  /* OOGL FORMAT: Nverts Nfaces Nedges */

  for (c = 0; c < count; c++) {
    fprintf(fichier,"%f %f %f\n",PX[c],PY[c],PZ[c]);
  }
  fprintf(fichier,"\n");

  for (c = 0; c < NbTri; c++) {
    if (PVALID[c]) {
      fprintf(fichier,"3 %d %d %d ",PTRI[c],PTRI[NbTri+c],PTRI[2*NbTri+c]);
      i1 = (int)PLIG[PTRI[c]];
      i2 = (int)PLIG[PTRI[NbTri+c]];
      i3 = (int)PLIG[PTRI[2*NbTri+c]];
      j1 = (int)PCOL[PTRI[c]];
      j2 = (int)PCOL[PTRI[NbTri+c]];
      j3 = (int)PCOL[PTRI[2*NbTri+c]];

      if (q_im_gauche == NULL)
	fcol = 0;
      else
	{
	  color = (int)(q_im_gauche[i1*nbcol+j1] + q_im_gauche[i2*nbcol+j2]
			+ q_im_gauche[i3*nbcol+j3])/3;
	  if (color) fcol = (float)color/255.0;
	  else fcol=0;
	}

      fprintf(fichier,"%f %f %f \n",fcol,fcol,fcol);
    }
  }
  fprintf(fichier,"\n");
  fclose (fichier);

  /********/
  /* FREE */
  /********/

  free(PTRI);
  free (PX);
  free (PY);
  free (PZ);
  free (PLIG);
  free (PCOL);
  free (PVALID);
  return (ERR_IM3D_OK);

}

